from __future__ import annotations

from asyncio import Future
from typing import Any, Callable, Dict, Tuple, TypedDict

from omu.client import Client
from omu.extension import Extension, ExtensionType
from omu.helper import Coro
from omu.network.bytebuffer import ByteReader, ByteWriter
from omu.network.packet import PacketType
from omu.serializer import Serializable, Serializer

from .endpoint import EndpointType

ENDPOINT_EXTENSION_TYPE = ExtensionType(
    "endpoint",
    lambda client: EndpointExtension(client),
    lambda: [],
)


class EndpointExtension(Extension):
    def __init__(self, client: Client) -> None:
        self.client = client
        self.response_promises: Dict[int, Future] = {}
        self.endpoints: Dict[str, Tuple[EndpointType, Coro[[Any], Any]]] = {}
        self.call_id = 0
        client.network.register_packet(
            ENDPOINT_CALL_PACKET,
            ENDPOINT_RECEIVE_PACKET,
            ENDPOINT_ERROR_PACKET,
        )
        client.network.add_packet_handler(ENDPOINT_RECEIVE_PACKET, self._on_receive)
        client.network.add_packet_handler(ENDPOINT_ERROR_PACKET, self._on_error)
        client.network.add_packet_handler(ENDPOINT_CALL_PACKET, self._on_call)
        client.network.listeners.connected += self.on_connected

    async def _on_receive(self, data: EndpointDataReq) -> None:
        if data["id"] not in self.response_promises:
            raise Exception(f"Received response for unknown call id {data['id']}")
        future = self.response_promises.pop(data["id"])
        future.set_result(data["data"])

    async def _on_error(self, data: EndpointError) -> None:
        if data["id"] not in self.response_promises:
            raise Exception(f"Received error for unknown call id {data['id']}")
        future = self.response_promises.pop(data["id"])
        future.set_exception(Exception(data["error"]))

    async def _on_call(self, data: EndpointDataReq) -> None:
        if data["type"] not in self.endpoints:
            return
        endpoint, func = self.endpoints[data["type"]]
        try:
            req = endpoint.request_serializer.deserialize(data["data"])
            res = await func(req)
            res_data = endpoint.response_serializer.serialize(res)
            await self.client.send(
                ENDPOINT_RECEIVE_PACKET,
                EndpointDataReq(type=data["type"], id=data["id"], data=res_data),
            )
        except Exception as e:
            await self.client.send(
                ENDPOINT_ERROR_PACKET,
                EndpointError(type=data["type"], id=data["id"], error=str(e)),
            )
            raise e

    async def on_connected(self) -> None:
        for endpoint, _ in self.endpoints.values():
            await self.client.send(ENDPOINT_REGISTER_PACKET, endpoint.identifier.key())

    def register[Req, Res](
        self, type: EndpointType[Req, Res], func: Coro[[Req], Res]
    ) -> None:
        if type.identifier.key() in self.endpoints:
            raise Exception(
                f"Endpoint for key {type.identifier.key()} already registered"
            )
        self.endpoints[type.identifier.key()] = (type, func)

    def bind[T, R](
        self,
        func: Coro[[T], R] | None = None,
        endpoint_type: EndpointType[T, R] | None = None,
    ):
        if endpoint_type is None:
            raise Exception("Endpoint type must be provided")

        def decorator(func: Coro[[T], R]) -> Coro[[T], R]:
            self.register(endpoint_type, func)
            return func

        if func:
            decorator(func)
        return decorator

    def listen(
        self, func: Coro | None = None, name: str | None = None
    ) -> Callable[[Coro], Coro]:
        def decorator(func: Coro) -> Coro:
            type = EndpointType.create_json(
                self.client.app.identifier,
                name or func.__name__,
            )
            self.register(type, func)
            return func

        if func:
            decorator(func)
        return decorator

    async def call[Req, Res](self, endpoint: EndpointType[Req, Res], data: Req) -> Res:
        try:
            self.call_id += 1
            future = Future[bytes]()
            self.response_promises[self.call_id] = future
            json = endpoint.request_serializer.serialize(data)
            await self.client.send(
                ENDPOINT_CALL_PACKET,
                EndpointDataReq(
                    type=endpoint.identifier.key(), id=self.call_id, data=json
                ),
            )
            res = await future
            return endpoint.response_serializer.deserialize(res)
        except Exception as e:
            raise Exception(
                f"Error calling endpoint {endpoint.identifier.key()}"
            ) from e


class EndpointReq(TypedDict):
    type: str
    id: int


class EndpointDataReq(EndpointReq):
    type: str
    id: int
    data: bytes


class EndpointError(EndpointReq):
    type: str
    id: int
    error: str


class CallSerializer(Serializable[EndpointDataReq, bytes]):
    def serialize(self, item: EndpointDataReq) -> bytes:
        writer = ByteWriter()
        writer.write_string(item["type"])
        writer.write_int(item["id"])
        writer.write_byte_array(item["data"])
        return writer.finish()

    def deserialize(self, item: bytes) -> EndpointDataReq:
        with ByteReader(item) as reader:
            type = reader.read_string()
            id = reader.read_int()
            data = reader.read_byte_array()
        return EndpointDataReq(type=type, id=id, data=data)


CALL_SERIALIZER = CallSerializer()

ENDPOINT_REGISTER_PACKET = PacketType[str].create_json(
    ENDPOINT_EXTENSION_TYPE,
    "register",
    Serializer.json(),
)
ENDPOINT_CALL_PACKET = PacketType[EndpointDataReq].create_serialized(
    ENDPOINT_EXTENSION_TYPE,
    "call",
    CALL_SERIALIZER,
)
ENDPOINT_RECEIVE_PACKET = PacketType[EndpointDataReq].create_serialized(
    ENDPOINT_EXTENSION_TYPE,
    "receive",
    CALL_SERIALIZER,
)
ENDPOINT_ERROR_PACKET = PacketType[EndpointError].create_json(
    ENDPOINT_EXTENSION_TYPE,
    "error",
)



from __future__ import annotations

import asyncio
from typing import Callable, List, Self

from omu.helper import Coro


class EventEmitter[**P]:
    def __init__(
        self,
        on_subscribe: Callable[[], None] | Coro[[], None] | None = None,
        on_empty: Callable[[], None] | Coro[[], None] | None = None,
    ) -> None:
        self.on_subscribe = on_subscribe
        self.on_empty = on_empty
        self._listeners: List[Coro[P, None]] = []

    @property
    def empty(self) -> bool:
        return len(self._listeners) == 0

    def subscribe(self, listener: Coro[P, None]) -> None:
        if listener in self._listeners:
            raise ValueError("Listener already subscribed")
        if self.on_subscribe and len(self._listeners) == 0:
            coroutine = self.on_subscribe()
            if asyncio.iscoroutine(coroutine):
                asyncio.create_task(coroutine)
        self._listeners.append(listener)

    def unsubscribe(self, listener: Coro[P, None]) -> None:
        if listener not in self._listeners:
            return
        self._listeners.remove(listener)
        if self.on_empty and len(self._listeners) == 0:
            coroutine = self.on_empty()
            if asyncio.iscoroutine(coroutine):
                asyncio.create_task(coroutine)

    def __iadd__(self, listener: Coro[P, None]) -> Self:
        self.subscribe(listener)
        return self

    def __isub__(self, listener: Coro[P, None]) -> Self:
        self.unsubscribe(listener)
        return self

    async def emit(self, *args: P.args, **kwargs: P.kwargs) -> None:
        for listener in self._listeners:
            await listener(*args, **kwargs)

    __call__ = emit



from omu.app import App
from omu.client import OmuClient
from omu.identifier import Identifier
from omu.network import Address
from omu.network.packet import PACKET_TYPES, Packet

address = Address(
    host="localhost",
    port=26423,
    secure=False,
)
IDENTIFIER = Identifier(
    "test",
    "test",
)
client = OmuClient(
    app=App(
        IDENTIFIER,
        version="0.0.1",
    ),
    address=address,
)


@client.network.listeners.connected.subscribe
async def on_connected() -> None:
    print("Connected")


@client.network.listeners.disconnected.subscribe
async def on_disconnected() -> None:
    print("Disconnected")


@client.network.listeners.packet.subscribe
async def on_event(event: Packet) -> None:
    print(event)


@client.network.add_packet_handler(PACKET_TYPES.READY)
async def on_ready(_) -> None:
    print("Ready")


if __name__ == "__main__":
    client.run()



from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING

from loguru import logger

from omu.extension.asset import ASSET_EXTENSION_TYPE, AssetExtension
from omu.extension.dashboard import (
    DASHBOARD_EXTENSION_TYPE,
    DashboardExtension,
)
from omu.extension.endpoint import (
    ENDPOINT_EXTENSION_TYPE,
    EndpointExtension,
)
from omu.extension.extension_manager import ExtensionManager
from omu.extension.message import (
    MESSAGE_EXTENSION_TYPE,
    MessageExtension,
)
from omu.extension.permission import PERMISSION_EXTENSION_TYPE, PermissionExtension
from omu.extension.registry import (
    REGISTRY_EXTENSION_TYPE,
    RegistryExtension,
)
from omu.extension.server import SERVER_EXTENSION_TYPE, ServerExtension
from omu.extension.table import TABLE_EXTENSION_TYPE, TableExtension
from omu.network import Address, Network
from omu.network.packet import Packet, PacketType
from omu.network.websocket_connection import WebsocketsConnection

from .client import Client, ClientListeners

if TYPE_CHECKING:
    from omu.app import App
    from omu.extension import ExtensionManager


class OmuClient(Client):
    def __init__(
        self,
        app: App,
        address: Address,
        connection: WebsocketsConnection | None = None,
        extension_registry: ExtensionManager | None = None,
        loop: asyncio.AbstractEventLoop | None = None,
    ):
        self._loop = loop or asyncio.get_event_loop()
        self._running = False
        self._listeners = ClientListeners()
        self._app = app
        self._network = Network(
            self,
            address,
            connection or WebsocketsConnection(self, address),
        )
        self._network.listeners.connected += self._listeners.ready.emit
        self._extensions = extension_registry or ExtensionManager(self)

        self._endpoints = self.extensions.register(ENDPOINT_EXTENSION_TYPE)
        self._tables = self.extensions.register(TABLE_EXTENSION_TYPE)
        self._registry = self.extensions.register(REGISTRY_EXTENSION_TYPE)
        self._message = self.extensions.register(MESSAGE_EXTENSION_TYPE)
        self._assets = self.extensions.register(ASSET_EXTENSION_TYPE)
        self._server = self.extensions.register(SERVER_EXTENSION_TYPE)
        self._permissions = self.extensions.register(PERMISSION_EXTENSION_TYPE)
        self._dashboard = self.extensions.register(DASHBOARD_EXTENSION_TYPE)

        self._loop.create_task(self._listeners.initialized.emit())

    @property
    def app(self) -> App:
        return self._app

    @property
    def loop(self) -> asyncio.AbstractEventLoop:
        return self._loop

    @property
    def network(self) -> Network:
        return self._network

    @property
    def extensions(self) -> ExtensionManager:
        return self._extensions

    @property
    def endpoints(self) -> EndpointExtension:
        return self._endpoints

    @property
    def tables(self) -> TableExtension:
        return self._tables

    @property
    def registry(self) -> RegistryExtension:
        return self._registry

    @property
    def message(self) -> MessageExtension:
        return self._message

    @property
    def assets(self) -> AssetExtension:
        return self._assets

    @property
    def server(self) -> ServerExtension:
        return self._server

    @property
    def permissions(self) -> PermissionExtension:
        return self._permissions

    @property
    def dashboard(self) -> DashboardExtension:
        return self._dashboard

    @property
    def running(self) -> bool:
        return self._running

    async def send[T](self, type: PacketType[T], data: T) -> None:
        await self._network.send(Packet(type, data))

    def run(self, *, token: str | None = None, reconnect: bool = True) -> None:
        try:
            self.loop.set_exception_handler(self.handle_exception)
            self.loop.create_task(self.start(token=token, reconnect=reconnect))
            self.loop.run_forever()
        finally:
            self.loop.close()
            asyncio.run(self.stop())

    def handle_exception(self, loop: asyncio.AbstractEventLoop, context: dict) -> None:
        logger.error(context["message"])
        exception = context.get("exception")
        if exception:
            raise exception

    async def start(self, *, token: str | None = None, reconnect: bool = True) -> None:
        if self._running:
            raise RuntimeError("Already running")
        self._running = True
        self.loop.create_task(self._network.connect(token=token, reconnect=reconnect))
        await self._listeners.started()

    async def stop(self) -> None:
        if not self._running:
            raise RuntimeError("Not running")
        self._running = False
        await self._network.disconnect()
        await self._listeners.stopped()

    @property
    def listeners(self) -> ClientListeners:
        return self._listeners



from __future__ import annotations

import abc
from dataclasses import dataclass
from typing import Callable

from omu.helper import Coro
from omu.identifier import Identifier
from omu.serializer import Serializable, Serializer


@dataclass(frozen=True)
class RegistryType[T]:
    identifier: Identifier
    default_value: T
    serializer: Serializable[T, bytes]

    @classmethod
    def create_json(
        cls,
        identifier: Identifier,
        name: str,
        default_value: T,
    ) -> RegistryType[T]:
        return cls(identifier / name, default_value, Serializer.json())

    @classmethod
    def create_serialized(
        cls,
        identifier: Identifier,
        name: str,
        default_value: T,
        serializer: Serializable[T, bytes],
    ) -> RegistryType[T]:
        return cls(identifier / name, default_value, serializer)


class Registry[T](abc.ABC):
    @abc.abstractmethod
    async def get(self) -> T: ...

    @abc.abstractmethod
    async def update(self, handler: Coro[[T], T]) -> None: ...

    @abc.abstractmethod
    def listen(self, handler: Coro[[T], None]) -> Callable[[], None]: ...



from __future__ import annotations

from dataclasses import dataclass
from typing import Any

from omu.identifier import Identifier
from omu.serializer import Serializable, Serializer


@dataclass(frozen=True)
class EndpointType[Req, Res]:
    identifier: Identifier
    request_serializer: Serializable[Req, bytes]
    response_serializer: Serializable[Res, bytes]

    @classmethod
    def create_json(
        cls,
        identifier: Identifier,
        name: str,
        request_serializer: Serializable[Req, Any] | None = None,
        response_serializer: Serializable[Res, Any] | None = None,
    ):
        return cls(
            identifier=identifier / name,
            request_serializer=Serializer.of(
                request_serializer or Serializer.noop()
            ).pipe(Serializer.json()),
            response_serializer=Serializer.of(
                response_serializer or Serializer.noop()
            ).pipe(Serializer.json()),
        )

    @classmethod
    def create_serialized(
        cls,
        identifier: Identifier,
        name: str,
        request_serializer: Serializable[Req, bytes],
        response_serializer: Serializable[Res, bytes],
    ):
        return cls(
            identifier=identifier / name,
            request_serializer=request_serializer,
            response_serializer=response_serializer,
        )



from __future__ import annotations

from typing import Final, List, NotRequired, TypedDict

from omu.identifier import Identifier
from omu.interface import Keyable
from omu.model import Model


class AppJson(TypedDict):
    identifier: str
    version: NotRequired[str] | None
    license: NotRequired[str] | None
    description: NotRequired[str] | None
    authors: NotRequired[List[str]] | None
    site_url: NotRequired[str] | None
    repository_url: NotRequired[str] | None
    image_url: NotRequired[str] | None


class App(Keyable, Model[AppJson]):
    def __init__(
        self,
        identifier: Identifier | str,
        *,
        version: str | None = None,
        license: str | None = None,
        description: str | None = None,
        authors: List[str] | None = None,
        site_url: str | None = None,
        repository_url: str | None = None,
        image_url: str | None = None,
    ) -> None:
        if isinstance(identifier, str):
            identifier = Identifier.from_key(identifier)
        self.identifier: Final[Identifier] = identifier
        self.name: Final[str] = "/".join(identifier.path)
        self.group: Final[str] = identifier.namespace
        self.version = version
        self.license = license
        self.description = description
        self.authors = authors
        self.site_url = site_url
        self.repository_url = repository_url
        self.image_url = image_url

    @classmethod
    def from_json(cls, json: AppJson) -> App:
        identifier = Identifier.from_key(json["identifier"])
        return cls(
            identifier,
            version=json.get("version"),
            license=json.get("license"),
            description=json.get("description"),
            authors=json.get("authors"),
            site_url=json.get("site_url"),
            repository_url=json.get("repository_url"),
            image_url=json.get("image_url"),
        )

    def to_json(self) -> AppJson:
        return AppJson(
            identifier=self.key(),
            version=self.version,
            license=self.license,
            description=self.description,
            authors=self.authors,
            site_url=self.site_url,
            repository_url=self.repository_url,
            image_url=self.image_url,
        )

    def key(self) -> str:
        return self.identifier.key()

    def __hash__(self) -> int:
        return hash(self.key())

    def __repr__(self) -> str:
        return f"App({self.key()})"



from __future__ import annotations

from typing import TypedDict

from omu.client import Client
from omu.extension import Extension, ExtensionType
from omu.extension.endpoint import EndpointType
from omu.identifier import Identifier
from omu.network.packet import PacketType
from omu.serializer import Serializer

from .dashboard import PermissionRequest

DASHBOARD_EXTENSION_TYPE = ExtensionType(
    "dashboard",
    lambda client: DashboardExtension(client),
    lambda: [],
)


class DashboardExtension(Extension):
    def __init__(self, client: Client):
        self.client = client

        self.client.network.register_packet(
            DASHBOARD_PERMISSION_REQUEST_PACKET,
            DASHBOARD_PERMISSION_ACCEPT_PACKET,
            DASHBOARD_PERMISSION_DENY_PACKET,
        )


class DashboardSetResponse(TypedDict):
    success: bool


DASHBOARD_SET_ENDPOINT = EndpointType[
    Identifier, DashboardSetResponse
].create_serialized(
    DASHBOARD_EXTENSION_TYPE,
    "set",
    request_serializer=Serializer.model(Identifier).pipe(Serializer.json()),
    response_serializer=Serializer.json(),
)
DASHBOARD_PERMISSION_REQUEST_PACKET = PacketType.create_json(
    DASHBOARD_EXTENSION_TYPE,
    "permission_request",
    Serializer.model(PermissionRequest),
)
DASHBOARD_PERMISSION_ACCEPT_PACKET = PacketType[int].create_json(
    DASHBOARD_EXTENSION_TYPE,
    "permission_accept",
)
DASHBOARD_PERMISSION_DENY_PACKET = PacketType[int].create_json(
    DASHBOARD_EXTENSION_TYPE,
    "permission_deny",
)



from __future__ import annotations

from dataclasses import dataclass
from typing import List, TypedDict

from omu.app import App, AppJson
from omu.extension.permission.permission import PermissionType, PermissionTypeJson
from omu.model import Model


class PermissionRequestJson(TypedDict):
    request_id: int
    app: AppJson
    permissions: List[PermissionTypeJson]


@dataclass(frozen=True)
class PermissionRequest(Model[PermissionRequestJson]):
    request_id: int
    app: App
    permissions: List[PermissionType]

    @classmethod
    def from_json(cls, json: PermissionRequestJson) -> PermissionRequest:
        return cls(
            request_id=json["request_id"],
            app=App.from_json(json["app"]),
            permissions=[PermissionType.from_json(p) for p in json["permissions"]],
        )

    def to_json(self) -> PermissionRequestJson:
        return {
            "request_id": self.request_id,
            "app": self.app.to_json(),
            "permissions": [p.to_json() for p in self.permissions],
        }



from typing import Dict, List, TypedDict

from omu.client import Client
from omu.extension import Extension, ExtensionType
from omu.extension.endpoint.endpoint import EndpointType
from omu.extension.permission.permission import PermissionType
from omu.network.packet.packet import PacketType

PLUGIN_EXTENSION_TYPE = ExtensionType(
    "plugin",
    lambda client: PluginExtension(client),
    lambda: [],
)


class PluginExtension(Extension):
    def __init__(self, client: Client):
        self.client = client
        self.plugins: Dict[str, str | None] = {}

        self.client.network.register_packet(
            PLUGIN_REQUIRE_PACKET,
        )
        self.client.network.listeners.connected += self.on_connected

    async def on_connected(self):
        await self.client.send(PLUGIN_REQUIRE_PACKET, self.plugins)
        await self.client.endpoints.call(PLUGIN_WAIT_ENDPOINT, [*self.plugins.keys()])

    def require(self, plugins: Dict[str, str | None]):
        self.plugins.update(plugins)
        self.client.permissions.require(PLUGIN_PERMISSION)


PLUGIN_PERMISSION = PermissionType.create(
    PLUGIN_EXTENSION_TYPE,
    "request",
)
PLUGIN_REQUIRE_PACKET = PacketType[Dict[str, str | None]].create_json(
    PLUGIN_EXTENSION_TYPE,
    "require",
)


class WaitResponse(TypedDict):
    success: bool


PLUGIN_WAIT_ENDPOINT = EndpointType[List[str], WaitResponse].create_json(
    PLUGIN_EXTENSION_TYPE,
    "wait",
)



from __future__ import annotations

from dataclasses import dataclass
from typing import TypedDict

from omu.identifier import Identifier
from omu.model import Model


class PermissionTypeJson(TypedDict):
    identifier: str


@dataclass(frozen=True)
class PermissionType(Model[PermissionTypeJson]):
    identifier: Identifier

    @classmethod
    def create(
        cls,
        identifier: Identifier,
        name: str,
    ) -> PermissionType:
        return PermissionType(
            identifier=identifier / name,
        )

    def to_json(self) -> PermissionTypeJson:
        return {
            "identifier": self.identifier.key(),
        }

    @classmethod
    def from_json(cls, json: PermissionTypeJson) -> PermissionType:
        return PermissionType(
            identifier=Identifier(json["identifier"]),
        )



from .app import App
from .client import Client, OmuClient
from .identifier import Identifier
from .network import Address, Network, NetworkStatus
from .plugin import Plugin

__all__ = [
    "Address",
    "Network",
    "NetworkStatus",
    "Client",
    "OmuClient",
    "App",
    "Identifier",
    "Plugin",
]



from typing import List, Mapping

from omu.client import Client
from omu.extension import Extension, ExtensionType
from omu.extension.endpoint import EndpointType
from omu.helper import instance
from omu.identifier import Identifier
from omu.network.bytebuffer import ByteReader, ByteWriter
from omu.serializer import Serializable, Serializer

ASSET_EXTENSION_TYPE = ExtensionType(
    "asset",
    lambda client: AssetExtension(client),
    lambda: [],
)
type Files = Mapping[Identifier, bytes]


@instance
class FILES_SERIALIZER(Serializable[Files, bytes]):
    def serialize(self, item: Files) -> bytes:
        writer = ByteWriter()
        writer.write_int(len(item))
        for identifier, value in item.items():
            writer.write_string(identifier.key())
            writer.write_byte_array(value)
        return writer.finish()

    def deserialize(self, item: bytes) -> Files:
        with ByteReader(item) as reader:
            count = reader.read_int()
            files: Files = {}
            for _ in range(count):
                identifier = Identifier.from_key(reader.read_string())
                value = reader.read_byte_array()
                files[identifier] = value
        return files


ASSET_UPLOAD_ENDPOINT = EndpointType[Files, List[Identifier]].create_serialized(
    ASSET_EXTENSION_TYPE,
    "upload",
    request_serializer=FILES_SERIALIZER,
    response_serializer=Serializer.model(Identifier).array().pipe(Serializer.json()),
)
ASSET_DOWNLOAD_ENDPOINT = EndpointType[List[Identifier], Files].create_serialized(
    ASSET_EXTENSION_TYPE,
    "download",
    request_serializer=Serializer.model(Identifier).array().pipe(Serializer.json()),
    response_serializer=FILES_SERIALIZER,
)


class AssetExtension(Extension):
    def __init__(self, client: Client) -> None:
        self.client = client

    async def upload(self, assets: Files) -> List[Identifier]:
        return await self.client.endpoints.call(ASSET_UPLOAD_ENDPOINT, assets)

    async def download(self, identifiers: List[Identifier]) -> Files:
        return await self.client.endpoints.call(ASSET_DOWNLOAD_ENDPOINT, identifiers)

    def url(self, identifier: Identifier) -> str:
        address = self.client.network.address
        protocol = "https" if address.secure else "http"
        return f"{protocol}://{address.host}:{address.port}/asset?id={identifier.key()}"

    def proxy(self, url: str) -> str:
        address = self.client.network.address
        protocol = "https" if address.secure else "http"
        return f"{protocol}://{address.host}:{address.port}/proxy?url={url}"



from __future__ import annotations

import asyncio
from dataclasses import dataclass, field
from typing import Dict, List, Literal

from omu.client import Client
from omu.event_emitter import EventEmitter
from omu.helper import Coro
from omu.identifier import Identifier

from .address import Address
from .connection import Connection
from .packet import Packet, PacketType
from .packet.packet_types import PACKET_TYPES, ConnectPacket
from .packet_mapper import PacketMapper


@dataclass(frozen=True)
class PacketListeners[T]:
    event_type: PacketType[T]
    listeners: EventEmitter[T] = field(default_factory=EventEmitter)


class Network:
    def __init__(self, client: Client, address: Address, connection: Connection):
        self._client = client
        self._address = address
        self._connection = connection
        self._connected = False
        self._listeners = NetworkListeners()
        self._tasks: List[Coro[[], None]] = []
        self._token: str | None = None
        self._closed_event = asyncio.Event()
        self._packet_mapper = PacketMapper()
        self._packet_handlers: Dict[Identifier, PacketListeners] = {}
        self._packet_mapper.register(
            PACKET_TYPES.CONNECT,
            PACKET_TYPES.DISCONNECT,
            PACKET_TYPES.TOKEN,
            PACKET_TYPES.READY,
        )

    @property
    def address(self) -> Address:
        return self._address

    def set_connection(self, connection: Connection) -> None:
        if self._connected:
            raise RuntimeError("Cannot change connection while connected")
        if self._connection:
            del self._connection
        self._connection = connection

    def register_packet(self, *packet_types: PacketType) -> None:
        self._packet_mapper.register(*packet_types)
        for packet_type in packet_types:
            if self._packet_handlers.get(packet_type.identifier):
                raise ValueError(
                    f"Event type {packet_type.identifier} already registered"
                )
            self._packet_handlers[packet_type.identifier] = PacketListeners(packet_type)

    def add_packet_handler[T](
        self,
        packet_type: PacketType[T],
        packet_handler: Coro[[T], None] | None = None,
    ):
        if not self._packet_handlers.get(packet_type.identifier):
            raise ValueError(f"Event type {packet_type.identifier} not registered")

        def decorator(packet_handler: Coro[[T], None]) -> None:
            self._packet_handlers[packet_type.identifier].listeners.subscribe(
                packet_handler
            )

        if packet_handler:
            decorator(packet_handler)
        return decorator

    @property
    def connected(self) -> bool:
        return self._connected

    async def connect(
        self, *, token: str | None = None, reconnect: bool = True
    ) -> None:
        if self._connected:
            raise RuntimeError("Already connected")

        self._token = token
        await self.disconnect()
        await self._connection.connect()
        self._connected = True
        await self.send(
            Packet(
                PACKET_TYPES.CONNECT,
                ConnectPacket(
                    app=self._client.app,
                    token=self._token,
                ),
            )
        )
        self._closed_event.clear()
        self._client.loop.create_task(self._listen())

        await self._listeners.status.emit("connected")
        await self._listeners.connected.emit()
        self._client.loop.create_task(self._dispatch_tasks())

        await self._closed_event.wait()

        if reconnect:
            await self.connect(token=self._token, reconnect=True)

    async def disconnect(self) -> None:
        if self._connection.closed:
            return
        self._connected = False
        await self._connection.close()
        self._closed_event.set()
        await self._listeners.status.emit("disconnected")
        await self._listeners.disconnected.emit()

    async def send(self, packet: Packet) -> None:
        if not self._connected:
            raise RuntimeError("Not connected")
        await self._connection.send(packet, self._packet_mapper)

    async def _listen(self) -> None:
        try:
            while not self._connection.closed:
                packet = await self._connection.receive(self._packet_mapper)
                self._client.loop.create_task(self.dispatch_packet(packet))
        finally:
            await self.disconnect()

    async def dispatch_packet(self, packet: Packet) -> None:
        await self._listeners.packet.emit(packet)
        packet_handler = self._packet_handlers.get(packet.type.identifier)
        if not packet_handler:
            return
        await packet_handler.listeners.emit(packet.data)

    @property
    def listeners(self) -> NetworkListeners:
        return self._listeners

    def add_task(self, task: Coro[[], None]) -> None:
        self._tasks.append(task)

    def remove_task(self, task: Coro[[], None]) -> None:
        self._tasks.remove(task)

    async def _dispatch_tasks(self) -> None:
        for task in self._tasks:
            await task()


type NetworkStatus = Literal["connecting", "connected", "disconnected"]


class NetworkListeners:
    def __init__(self) -> None:
        self.connected = EventEmitter()
        self.disconnected = EventEmitter()
        self.packet = EventEmitter[Packet]()
        self.status = EventEmitter[NetworkStatus]()



from .dashboard import PermissionRequest
from .dashboard_extension import DASHBOARD_EXTENSION_TYPE, DashboardExtension

__all__ = [
    "PermissionRequest",
    "DashboardExtension",
    "DASHBOARD_EXTENSION_TYPE",
]



import abc
from dataclasses import dataclass
from typing import Callable

from omu.helper import Coro
from omu.identifier import Identifier
from omu.serializer import Serializable, Serializer


@dataclass(frozen=True)
class MessageType[T]:
    identifier: Identifier
    serializer: Serializable[T, bytes]

    @classmethod
    def create_json(
        cls,
        identifier: Identifier,
        name: str,
    ):
        return cls(
            identifier=identifier / name,
            serializer=Serializer.json(),
        )

    @classmethod
    def create_serialized(
        cls,
        identifier: Identifier,
        name: str,
        serializer: Serializable[T, bytes],
    ):
        return cls(
            identifier=identifier / name,
            serializer=serializer,
        )


class Message[T](abc.ABC):
    @abc.abstractmethod
    def listen(self, listener: Coro[[T], None]) -> Callable[[], None]: ...

    @abc.abstractmethod
    async def broadcast(self, body: T) -> None: ...



from typing import Dict, List

from omu.client import Client
from omu.extension import Extension, ExtensionType
from omu.extension.endpoint.endpoint import EndpointType
from omu.identifier import Identifier
from omu.network.packet.packet import PacketType
from omu.serializer import Serializer

from .permission import PermissionType

PERMISSION_EXTENSION_TYPE = ExtensionType(
    "permission",
    lambda client: PermissionExtension(client),
    lambda: [],
)


class PermissionExtension(Extension):
    def __init__(self, client: Client):
        self.client = client
        self.permissions: List[PermissionType] = []
        self.registered_permissions: Dict[Identifier, PermissionType] = {}
        self.required_permissions: Dict[Identifier, PermissionType] = {}
        client.network.register_packet(
            PERMISSION_REGISTER_PACKET,
            PERMISSION_GRANT_PACKET,
        )
        client.network.add_packet_handler(
            PERMISSION_GRANT_PACKET,
            self.handle_grant,
        )
        client.network.listeners.connected += self.on_connected

    def register(self, permission: PermissionType):
        base_identifier = self.client.app.identifier
        if not permission.identifier.is_subpath_of(base_identifier):
            raise ValueError(
                f"Permission identifier {permission.identifier} is not a subpath of app identifier {base_identifier}"
            )
        self.registered_permissions[permission.identifier] = permission

    def require(self, permission: PermissionType):
        self.required_permissions[permission.identifier] = permission

    def has(self, permission_identifier: Identifier):
        return permission_identifier in self.permissions

    async def on_connected(self):
        await self.client.send(
            PERMISSION_REGISTER_PACKET,
            [*self.registered_permissions.values()],
        )
        await self.client.endpoints.call(
            PERMISSION_REQUEST_ENDPOINT,
            [*self.required_permissions.keys()],
        )

    async def handle_grant(self, permissions: List[PermissionType]):
        self.permissions = permissions


PERMISSION_REGISTER_PACKET = PacketType.create_json(
    PERMISSION_EXTENSION_TYPE,
    "register",
    Serializer.model(PermissionType).array(),
)
PERMISSION_REQUEST_ENDPOINT = EndpointType[List[Identifier], None].create_json(
    PERMISSION_EXTENSION_TYPE,
    "request",
    request_serializer=Serializer.model(Identifier).array(),
)
PERMISSION_GRANT_PACKET = PacketType.create_json(
    PERMISSION_EXTENSION_TYPE,
    "grant",
    Serializer.model(PermissionType).array(),
)



from __future__ import annotations

from typing import Mapping

from omu.app import App
from omu.identifier import Identifier
from omu.model import Model
from omu.serializer import Serializer

from .packet import PacketType


class ConnectPacket(Model):
    def __init__(self, app: App, token: str | None = None):
        self.app = app
        self.token = token

    def to_json(self) -> Mapping:
        return {
            "app": self.app.to_json(),
            "token": self.token,
        }

    @classmethod
    def from_json(cls, json: Mapping) -> ConnectPacket:
        return cls(
            app=App.from_json(json["app"]),
            token=json["token"],
        )


class DisconnectPacket(Model):
    def __init__(self, reason: str):
        self.reason = reason

    def to_json(self) -> Mapping:
        return {"reason": self.reason}

    @classmethod
    def from_json(cls, json: Mapping) -> DisconnectPacket:
        return cls(
            reason=json["reason"],
        )


class PACKET_TYPES:
    IDENTIFIER = Identifier("core", "packet")
    CONNECT = PacketType.create_json(
        IDENTIFIER,
        "connect",
        Serializer.model(ConnectPacket),
    )
    DISCONNECT = PacketType.create_json(
        IDENTIFIER,
        "disconnect",
        Serializer.model(DisconnectPacket),
    )
    TOKEN = PacketType[str].create_json(
        IDENTIFIER,
        "token",
        Serializer.noop(),
    )
    READY = PacketType[None].create_json(
        IDENTIFIER,
        "ready",
        Serializer.noop(),
    )



from __future__ import annotations

from dataclasses import dataclass
from typing import Callable, List

from omu.client import Client
from omu.extension import Extension, ExtensionType
from omu.extension.endpoint import EndpointType
from omu.helper import Coro
from omu.identifier import Identifier
from omu.network.bytebuffer import ByteReader, ByteWriter
from omu.network.packet import PacketType
from omu.serializer import Serializable, Serializer

from .registry import Registry, RegistryType

REGISTRY_EXTENSION_TYPE = ExtensionType(
    "registry",
    lambda client: RegistryExtension(client),
    lambda: [],
)


@dataclass(frozen=True)
class RegistryData:
    key: str
    existing: bool
    value: bytes


class RegistryDataSerializer(Serializable[RegistryData, bytes]):
    def serialize(self, item: RegistryData) -> bytes:
        writer = ByteWriter()
        writer.write_string(item.key)
        writer.write_boolean(item.existing)
        writer.write_byte_array(item.value)
        return writer.finish()

    def deserialize(self, item: bytes) -> RegistryData:
        with ByteReader(item) as reader:
            key = reader.read_string()
            existing = reader.read_boolean()
            value = reader.read_byte_array()
            return RegistryData(key, existing, value)


REGISTRY_UPDATE_PACKET = PacketType[RegistryData].create_serialized(
    REGISTRY_EXTENSION_TYPE,
    "update",
    serializer=RegistryDataSerializer(),
)
REGISTRY_LISTEN_PACKET = PacketType[str].create_json(REGISTRY_EXTENSION_TYPE, "listen")
REGISTRY_GET_ENDPOINT = EndpointType[str, RegistryData].create_serialized(
    REGISTRY_EXTENSION_TYPE,
    "get",
    request_serializer=Serializer.json(),
    response_serializer=RegistryDataSerializer(),
)


class RegistryExtension(Extension):
    def __init__(self, client: Client) -> None:
        self.client = client
        client.network.register_packet(REGISTRY_UPDATE_PACKET)

    def get[T](self, registry_type: RegistryType[T]) -> Registry[T]:
        return RegistryImpl(
            self.client,
            registry_type.identifier,
            registry_type.default_value,
            registry_type.serializer,
        )

    def create[T](self, name: str, default_value: T) -> Registry[T]:
        identifier = self.client.app.identifier / name
        return RegistryImpl(self.client, identifier, default_value, Serializer.json())


class RegistryImpl[T](Registry[T]):
    def __init__(
        self,
        client: Client,
        identifier: Identifier,
        default_value: T,
        serializer: Serializable[T, bytes],
    ) -> None:
        self.client = client
        self.identifier = identifier
        self.default_value = default_value
        self.serializer = serializer
        self.key = identifier.key()
        self.listeners: List[Coro[[T], None]] = []
        self.listening = False
        client.network.add_packet_handler(REGISTRY_UPDATE_PACKET, self._on_update)

    async def get(self) -> T:
        result = await self.client.endpoints.call(REGISTRY_GET_ENDPOINT, self.key)
        if not result.existing:
            return self.default_value
        return self.serializer.deserialize(result.value)

    async def update(self, handler: Coro[[T], T]) -> None:
        value = await self.get()
        new_value = await handler(value)
        await self.client.send(
            REGISTRY_UPDATE_PACKET,
            RegistryData(
                self.key,
                True,
                self.serializer.serialize(new_value),
            ),
        )

    def listen(self, handler: Coro[[T], None]) -> Callable[[], None]:
        if not self.listening:
            self.client.network.add_task(
                lambda: self.client.send(REGISTRY_LISTEN_PACKET, self.key)
            )
            self.listening = True

        self.listeners.append(handler)
        return lambda: self.listeners.remove(handler)

    async def _on_update(self, event: RegistryData) -> None:
        if event.key != self.key:
            return
        if event.existing:
            value = self.serializer.deserialize(event.value)
        else:
            value = self.default_value
        for listener in self.listeners:
            await listener(value)



from .address import Address
from .network import Network, NetworkStatus
from .packet import Packet, PacketType

__all__ = [
    "Address",
    "Network",
    "NetworkStatus",
    "Network",
    "Packet",
    "PacketType",
]



from omu.app import App
from omu.client import Client
from omu.extension import Extension, ExtensionType
from omu.extension.endpoint import EndpointType
from omu.extension.table import TABLE_EXTENSION_TYPE, TableType

SERVER_EXTENSION_TYPE = ExtensionType(
    "server", lambda client: ServerExtension(client), lambda: []
)

APPS_TABLE_TYPE = TableType.create_model(
    SERVER_EXTENSION_TYPE,
    "apps",
    App,
)
SHUTDOWN_ENDPOINT_TYPE = EndpointType[bool, bool].create_json(
    SERVER_EXTENSION_TYPE,
    "shutdown",
)


class ServerExtension(Extension):
    def __init__(self, client: Client) -> None:
        self.client = client
        tables = client.extensions.get(TABLE_EXTENSION_TYPE)
        self.apps = tables.get(APPS_TABLE_TYPE)

    async def shutdown(self, restart: bool = False) -> bool:
        return await self.client.endpoints.call(SHUTDOWN_ENDPOINT_TYPE, restart)



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

from omu.identifier import Identifier
from omu.serializer import Serializer

if TYPE_CHECKING:
    from omu.serializer import Serializable


@dataclass(frozen=True)
class PacketData:
    type: str
    data: bytes


@dataclass(frozen=True)
class Packet[T]:
    type: PacketType[T]
    data: T


@dataclass(frozen=True)
class PacketType[T]:
    identifier: Identifier
    serializer: Serializable[T, bytes]

    @classmethod
    def create_json[_T](
        cls,
        identifier: Identifier,
        name: str,
        serializer: Serializable[_T, Any] | None = None,
    ) -> PacketType[_T]:
        return PacketType(
            identifier=identifier / name,
            serializer=Serializer.of(serializer or Serializer.noop()).pipe(
                Serializer.json()
            ),
        )

    @classmethod
    def create_serialized[_T](
        cls,
        identifier: Identifier,
        name: str,
        serializer: Serializable[_T, bytes],
    ) -> PacketType[_T]:
        return PacketType(
            identifier=identifier / name,
            serializer=serializer,
        )



from __future__ import annotations

from dataclasses import dataclass
from typing import Callable, List

from omu.client import Client
from omu.extension import Extension, ExtensionType
from omu.helper import Coro
from omu.identifier import Identifier
from omu.network.bytebuffer import ByteReader, ByteWriter
from omu.network.packet import PacketType
from omu.serializer import Serializable

from .message import Message, MessageType

MESSAGE_EXTENSION_TYPE = ExtensionType(
    "message",
    lambda client: MessageExtension(client),
    lambda: [],
)


@dataclass
class MessageData:
    key: str
    body: bytes


class MessageSerializer(Serializable[MessageData, bytes]):
    def serialize(self, item: MessageData) -> bytes:
        writer = ByteWriter()
        writer.write_string(item.key)
        writer.write_byte_array(item.body)
        return writer.finish()

    def deserialize(self, item: bytes) -> MessageData:
        with ByteReader(item) as reader:
            key = reader.read_string()
            body = reader.read_byte_array()
        return MessageData(key=key, body=body)


MESSAGE_LISTEN_PACKET = PacketType[str].create_json(MESSAGE_EXTENSION_TYPE, "listen")
MESSAGE_BROADCAST_PACKET = PacketType[MessageData].create_serialized(
    MESSAGE_EXTENSION_TYPE,
    "broadcast",
    MessageSerializer(),
)


class MessageExtension(Extension):
    def __init__(self, client: Client):
        self.client = client
        self._message_identifiers: List[Identifier] = []
        client.network.register_packet(
            MESSAGE_LISTEN_PACKET,
            MESSAGE_BROADCAST_PACKET,
        )

    def create[T](self, name: str, _t: type[T] | None = None) -> Message[T]:
        identifier = self.client.app.identifier / name
        if identifier in self._message_identifiers:
            raise Exception(f"Message {identifier} already exists")
        self._message_identifiers.append(identifier)
        type = MessageType.create_json(identifier, name)
        return MessageImpl(self.client, type)

    def get[T](self, message_type: MessageType[T]) -> Message[T]:
        return MessageImpl(self.client, message_type)


class MessageImpl[T](Message):
    def __init__(self, client: Client, message_type: MessageType[T]):
        self.client = client
        self.key = message_type.identifier.key()
        self.serializer = message_type.serializer
        self.listeners = []
        self.listening = False
        client.network.add_packet_handler(MESSAGE_BROADCAST_PACKET, self._on_broadcast)

    async def broadcast(self, body: T) -> None:
        data = self.serializer.serialize(body)
        await self.client.send(
            MESSAGE_BROADCAST_PACKET,
            MessageData(key=self.key, body=data),
        )

    def listen(self, listener: Coro[[T], None]) -> Callable[[], None]:
        self.listeners.append(listener)
        if not self.listening:
            self.client.network.add_task(self._listen)
            self.listening = True
        return lambda: self.listeners.remove(listener)

    async def _listen(self) -> None:
        await self.client.send(MESSAGE_LISTEN_PACKET, self.key)

    async def _on_broadcast(self, data: MessageData) -> None:
        if data.key != self.key:
            return
        for listener in self.listeners:
            await listener(data.body)



from typing import (
    AsyncGenerator,
    Callable,
    Dict,
    Iterable,
    List,
    Mapping,
    TypedDict,
)

from omu.client import Client
from omu.extension import Extension, ExtensionType
from omu.extension.endpoint import EndpointType
from omu.helper import AsyncCallback, Coro
from omu.identifier import Identifier
from omu.interface import Keyable
from omu.network.bytebuffer import ByteReader, ByteWriter
from omu.network.packet import PacketType
from omu.serializer import JsonSerializable, Serializable, Serializer

from .table import (
    Table,
    TableConfig,
    TableListeners,
    TableType,
)

type ModelType[T: Keyable, D] = JsonSerializable[T, D]


class TableExtension(Extension):
    def __init__(self, client: Client):
        self._client = client
        self._tables: Dict[Identifier, Table] = {}
        client.network.register_packet(
            TABLE_CONFIG_SET_PACKET,
            TABLE_LISTEN_PACKET,
            TABLE_PROXY_LISTEN_PACKET,
            TABLE_PROXY_PACKET,
            TABLE_ITEM_ADD_PACKET,
            TABLE_ITEM_UPDATE_PACKET,
            TABLE_ITEM_REMOVE_EVENT,
            TABLE_ITEM_CLEAR_PACKET,
        )

    def create[T](
        self,
        identifier: Identifier,
        serializer: Serializable[T, bytes],
        key_function: Callable[[T], str],
    ) -> Table[T]:
        if self.has(identifier):
            raise ValueError(f"Table with identifier {identifier} already exists")
        table = TableImpl(
            self._client,
            identifier=identifier,
            serializer=serializer,
            key_function=key_function,
        )
        self._tables[identifier] = table
        return table

    def get[T](self, type: TableType[T]) -> Table[T]:
        if self.has(type.identifier):
            return self._tables[type.identifier]
        return self.create(type.identifier, type.serializer, type.key_func)

    def model[T: Keyable, D](
        self, identifier: Identifier, name: str, type: type[ModelType[T, D]]
    ) -> Table[T]:
        identifier = identifier / name
        if self.has(identifier):
            return self._tables[identifier]
        return self.create(
            identifier,
            Serializer.model(type).pipe(Serializer.json()),
            lambda item: item.key(),
        )

    def has(self, identifier: Identifier) -> bool:
        return identifier in self._tables


TABLE_EXTENSION_TYPE = ExtensionType(
    "table", lambda client: TableExtension(client), lambda: []
)


class TableEventData(TypedDict):
    type: str


class TableItemsData(TableEventData):
    items: Dict[str, bytes]


class TableKeysData(TableEventData):
    keys: List[str]


class TableProxyData(TableItemsData):
    key: int


class TableItemsSerielizer(Serializable[TableItemsData, bytes]):
    def serialize(self, item: TableItemsData) -> bytes:
        writer = ByteWriter()
        writer.write_string(item["type"])
        writer.write_int(len(item["items"]))
        for key, value in item["items"].items():
            writer.write_string(key)
            writer.write_byte_array(value)
        return writer.finish()

    def deserialize(self, item: bytes) -> TableItemsData:
        with ByteReader(item) as reader:
            type = reader.read_string()
            item_count = reader.read_int()
            items: Mapping[str, bytes] = {}
            for _ in range(item_count):
                key = reader.read_string()
                value = reader.read_byte_array()
                items[key] = value
        return {"type": type, "items": items}


class TableProxySerielizer(Serializable[TableProxyData, bytes]):
    def serialize(self, item: TableProxyData) -> bytes:
        writer = ByteWriter()
        writer.write_string(item["type"])
        writer.write_int(item["key"])
        writer.write_int(len(item["items"]))
        for key, value in item["items"].items():
            writer.write_string(key)
            writer.write_byte_array(value)
        return writer.finish()

    def deserialize(self, item: bytes) -> TableProxyData:
        with ByteReader(item) as reader:
            type = reader.read_string()
            key = reader.read_int()
            item_count = reader.read_int()
            items: Dict[str, bytes] = {}
            for _ in range(item_count):
                item_key = reader.read_string()
                value = reader.read_byte_array()
                items[item_key] = value
        return {"type": type, "key": key, "items": items}


class SetConfigReq(TypedDict):
    type: str
    config: TableConfig


TABLE_CONFIG_SET_PACKET = PacketType[SetConfigReq].create_json(
    TABLE_EXTENSION_TYPE, "config_set"
)
TABLE_LISTEN_PACKET = PacketType[str].create_json(TABLE_EXTENSION_TYPE, name="listen")
TABLE_PROXY_LISTEN_PACKET = PacketType[str].create_json(
    TABLE_EXTENSION_TYPE, "proxy_listen"
)
TABLE_PROXY_PACKET = PacketType[TableProxyData].create_serialized(
    TABLE_EXTENSION_TYPE,
    "proxy",
    serializer=TableProxySerielizer(),
)
TABLE_PROXY_ENDPOINT = EndpointType[TableProxyData, int].create_serialized(
    TABLE_EXTENSION_TYPE,
    "proxy",
    request_serializer=TableProxySerielizer(),
    response_serializer=Serializer.json(),
)
TABLE_ITEM_ADD_PACKET = PacketType[TableItemsData].create_serialized(
    TABLE_EXTENSION_TYPE, "item_add", TableItemsSerielizer()
)
TABLE_ITEM_UPDATE_PACKET = PacketType[TableItemsData].create_serialized(
    TABLE_EXTENSION_TYPE, "item_update", TableItemsSerielizer()
)
TABLE_ITEM_REMOVE_EVENT = PacketType[TableItemsData].create_serialized(
    TABLE_EXTENSION_TYPE, "item_remove", TableItemsSerielizer()
)
TABLE_ITEM_CLEAR_PACKET = PacketType[TableEventData].create_json(
    TABLE_EXTENSION_TYPE, "item_clear"
)
TABLE_ITEM_GET_ENDPOINT = EndpointType[TableKeysData, TableItemsData].create_serialized(
    TABLE_EXTENSION_TYPE,
    "item_get",
    request_serializer=Serializer.json(),
    response_serializer=TableItemsSerielizer(),
)


class TableFetchReq(TypedDict):
    type: str
    before: int | None
    after: int | None
    cursor: str | None


TABLE_ITEM_FETCH_ENDPOINT = EndpointType[
    TableFetchReq, TableItemsData
].create_serialized(
    TABLE_EXTENSION_TYPE,
    "item_fetch",
    request_serializer=Serializer.json(),
    response_serializer=TableItemsSerielizer(),
)
TABLE_ITEM_FETCH_ALL_ENDPOINT = EndpointType[
    TableEventData, TableItemsData
].create_serialized(
    TABLE_EXTENSION_TYPE,
    "item_fetch_all",
    request_serializer=Serializer.json(),
    response_serializer=TableItemsSerielizer(),
)
TABLE_ITEM_SIZE_ENDPOINT = EndpointType[TableEventData, int].create_json(
    TABLE_EXTENSION_TYPE, "item_size"
)


class TableImpl[T](Table[T]):
    def __init__(
        self,
        client: Client,
        identifier: Identifier,
        serializer: Serializable[T, bytes],
        key_function: Callable[[T], str],
    ):
        self._client = client
        self._identifier = identifier
        self._serializer = serializer
        self._key_function = key_function
        self._cache: Dict[str, T] = {}
        self._listeners = TableListeners[T](self)
        self._proxies: List[Coro[[T], T | None]] = []
        self._chunk_size = 100
        self._cache_size: int | None = None
        self._listening = False
        self._config: TableConfig | None = None
        self.key = identifier.key()

        client.network.add_packet_handler(TABLE_PROXY_PACKET, self._on_proxy)
        client.network.add_packet_handler(TABLE_ITEM_ADD_PACKET, self._on_item_add)
        client.network.add_packet_handler(
            TABLE_ITEM_UPDATE_PACKET, self._on_item_update
        )
        client.network.add_packet_handler(TABLE_ITEM_REMOVE_EVENT, self._on_item_remove)
        client.network.add_packet_handler(TABLE_ITEM_CLEAR_PACKET, self._on_item_clear)
        client.network.add_task(self.on_connected)

    @property
    def cache(self) -> Mapping[str, T]:
        return self._cache

    async def get(self, key: str) -> T | None:
        if key in self._cache:
            return self._cache[key]
        res = await self._client.endpoints.call(
            TABLE_ITEM_GET_ENDPOINT, TableKeysData(type=self.key, keys=[key])
        )
        items = self._parse_items(res["items"])
        self._cache.update(items)
        if key in items:
            return items[key]
        return None

    async def add(self, *items: T) -> None:
        data = self._serialize_items(items)
        await self._client.send(
            TABLE_ITEM_ADD_PACKET, TableItemsData(type=self.key, items=data)
        )

    async def update(self, *items: T) -> None:
        data = self._serialize_items(items)
        await self._client.send(
            TABLE_ITEM_UPDATE_PACKET, TableItemsData(type=self.key, items=data)
        )

    async def remove(self, *items: T) -> None:
        data = self._serialize_items(items)
        await self._client.send(
            TABLE_ITEM_REMOVE_EVENT, TableItemsData(type=self.key, items=data)
        )

    async def clear(self) -> None:
        await self._client.send(TABLE_ITEM_CLEAR_PACKET, TableEventData(type=self.key))

    async def fetch_items(
        self,
        before: int | None = None,
        after: int | None = None,
        cursor: str | None = None,
    ) -> Dict[str, T]:
        items_response = await self._client.endpoints.call(
            TABLE_ITEM_FETCH_ENDPOINT,
            TableFetchReq(type=self.key, before=before, after=after, cursor=cursor),
        )
        items = self._parse_items(items_response["items"])
        await self.update_cache(items)
        return items

    async def fetch_all(self) -> Dict[str, T]:
        items_response = await self._client.endpoints.call(
            TABLE_ITEM_FETCH_ALL_ENDPOINT, TableEventData(type=self.key)
        )
        items = self._parse_items(items_response["items"])
        await self.update_cache(items)
        return items

    async def iterate(
        self,
        backward: bool = False,
        cursor: str | None = None,
    ) -> AsyncGenerator[T, None]:
        items = await self.fetch_items(
            before=self._chunk_size if backward else None,
            after=self._chunk_size if not backward else None,
            cursor=cursor,
        )
        for item in items.values():
            yield item
        while len(items) > 0:
            cursor = next(iter(items.keys()))
            items = await self.fetch_items(
                before=self._chunk_size if backward else None,
                after=self._chunk_size if not backward else None,
                cursor=cursor,
            )
            for item in items.values():
                yield item
            items.pop(cursor, None)

    async def size(self) -> int:
        res = await self._client.endpoints.call(
            TABLE_ITEM_SIZE_ENDPOINT, TableEventData(type=self.key)
        )
        return res

    def listen(
        self, listener: AsyncCallback[Mapping[str, T]] | None = None
    ) -> Callable[[], None]:
        self._listening = True
        if listener is not None:
            self._listeners.cache_update += listener
            return lambda: self._listeners.cache_update.unsubscribe(listener)
        return lambda: None

    def proxy(self, callback: Coro[[T], T | None]) -> Callable[[], None]:
        self._proxies.append(callback)
        return lambda: self._proxies.remove(callback)

    def set_config(self, config: TableConfig) -> None:
        self._config = config

    async def on_connected(self) -> None:
        if self._config is not None:
            await self._client.send(
                TABLE_CONFIG_SET_PACKET,
                SetConfigReq(type=self.key, config=self._config),
            )
        if self._listening:
            await self._client.send(TABLE_LISTEN_PACKET, self.key)
        if len(self._proxies) > 0:
            await self._client.send(TABLE_PROXY_LISTEN_PACKET, self.key)

    async def _on_proxy(self, event: TableProxyData) -> None:
        if event["type"] != self.key:
            return
        items = self._parse_items(event["items"])
        for proxy in self._proxies:
            for key, item in list(items.items()):
                updated_item = await proxy(item)
                if updated_item is None:
                    del items[key]
                else:
                    items[key] = updated_item
        serialized_items = self._serialize_items(items.values())
        await self._client.endpoints.call(
            TABLE_PROXY_ENDPOINT,
            TableProxyData(
                type=self.key,
                key=event["key"],
                items=serialized_items,
            ),
        )

    async def _on_item_add(self, event: TableItemsData) -> None:
        if event["type"] != self.key:
            return
        items = self._parse_items(event["items"])
        await self._listeners.add(items)
        await self.update_cache(items)

    async def _on_item_update(self, event: TableItemsData) -> None:
        if event["type"] != self.key:
            return
        items = self._parse_items(event["items"])
        await self._listeners.update(items)
        await self.update_cache(items)

    async def _on_item_remove(self, event: TableItemsData) -> None:
        if event["type"] != self.key:
            return
        items = self._parse_items(event["items"])
        await self._listeners.remove(items)
        for key in items.keys():
            if key not in self._cache:
                continue
            del self._cache[key]
        await self._listeners.cache_update(self._cache)

    async def _on_item_clear(self, event: TableEventData) -> None:
        if event["type"] != self.key:
            return
        await self._listeners.clear()
        self._cache.clear()
        await self._listeners.cache_update(self._cache)

    async def update_cache(self, items: Mapping[str, T]) -> None:
        if self._cache_size is None:
            self._cache = {**items}
        else:
            merged_cache = {**self._cache, **items}
            cache_array = tuple(merged_cache.items())
            self._cache = dict(cache_array[: self._cache_size])
        await self._listeners.cache_update(self._cache)

    def _parse_items(self, items: Mapping[str, bytes]) -> Dict[str, T]:
        parsed_items: Mapping[str, T] = {}
        for key, item_bytes in items.items():
            item = self._serializer.deserialize(item_bytes)
            if item is None:
                raise ValueError(f"Failed to deserialize item with key: {key}")
            parsed_items[key] = item
        return parsed_items

    def _serialize_items(self, items: Iterable[T]) -> Dict[str, bytes]:
        serialized_items: Mapping[str, bytes] = {}
        for item in items:
            key = self._key_function(item)
            serialized_items[key] = self._serializer.serialize(item)
        return serialized_items

    def set_cache_size(self, size: int | None) -> None:
        self._cache_size = size

    @property
    def listeners(self) -> TableListeners[T]:
        return self._listeners
